use criterion::{Criterion, BenchmarkId, criterion_group, criterion_main,
                BatchSize, PlotConfiguration, AxisScale, SamplingMode};

mod support;

const SLX_PATH: &'static str = "C:/Wolverine/SLX/sse.exe";
const SEED:  u64 = 100000;
const RANGE: u32 =     10;
const STEP: Time =   1000.0;

fn ferry_bench(c: &mut Criterion) {
	let mut group = c.benchmark_group("Ferry");
	
	// set-up the benchmark parameters
	group.confidence_level(0.99);
	group.plot_config(
		PlotConfiguration::default()
			.summary_scale(AxisScale::Logarithmic)
	);
	group.sampling_mode(SamplingMode::Linear);
	
	// vary in the length of the simulation run
	for (i, sim_duration) in (0..RANGE).map(|c| Time::from(1 << c)*STEP).enumerate() {
		let duration = sim_duration.to_string();
		let args = [
			"/silent",
			"/stdout",
			"/noicon",
			"/nowarn",
			"/noxwarn",
			"/#BENCH",
			"slx\\ferry.slx",
			duration.as_str()
		];
		
		// benchmark the SLX implementation
		group.bench_function(
			BenchmarkId::new("SLX", sim_duration),
			|b| b.iter_custom(|iters|
				support::slx_bench(
					SLX_PATH,
					&args,
					iters as usize
				).expect("couldn't benchmark the SLX program")
			)
		);
		
		// benchmark the Rust implementation
		group.bench_function(
			BenchmarkId::new("Rust", sim_duration),
			|b| b.iter_batched(
				|| Shared {
					master_rng: RefCell::new(SmallRng::seed_from_u64(SEED*((i+1) as u64))),
					ferry_cargo_len: RandomVar::new(),
					ferry_load_time: RandomVar::new(),
					car_wait_time: RandomVar::new(),
				},
				|shared| simulation(
					shared,
					|sim| sim.process(ferry(sim, sim_duration, 2, 4))
				),
				BatchSize::SmallInput
			)
		);
	}
	
	group.finish();
}

criterion_group!(benches, ferry_bench);
criterion_main!(benches);

/* ****************************** Ferry Example ***************************** */
use simcore_rs::{Time, SimContext, Sender, Receiver, RandomVar, channel, simulation};
use rand::{rngs::SmallRng, SeedableRng, Rng};
use rand_distr::{Exp, Normal, Distribution};
use std::cell::RefCell;

const HARBOR_DISTANCE: Time  = 10.0;
const FERRY_TIMEOUT  : Time  = 5.0;
const FERRY_CAPACITY : usize = 5;

struct Shared {
	master_rng: RefCell<SmallRng>,
	ferry_cargo_len: RandomVar,
	ferry_load_time: RandomVar,
	car_wait_time: RandomVar,
}

#[derive(Debug)]
struct Car {
	arrival_time: Time,
	load_duration: Time
}

struct Pier {
	rng: SmallRng,
	landing_site: Sender<Car>
}

struct Ferry {
	cargo: Vec<Car>,
	timeout: Time,
	travel_time: Time,
	piers: Vec<Receiver<Car>>
}

impl Pier {
	async fn actions(mut self, sim: SimContext<'_,Shared>) {
		let arrival_delay = Exp::new(0.1).unwrap();
		let loading_delay = Normal::new(0.5, 0.2).unwrap();
		
		loop {
			sim.advance(arrival_delay.sample(&mut self.rng)).await;
			self.landing_site.send(Car {
				arrival_time: sim.now(),
				load_duration: loading_delay
					.sample_iter(&mut self.rng)
					.find(|&val| val >= 0.0)
					.unwrap()
			}).await.expect("no ferries in the simulation");
		}
	}
}

impl Ferry {
	async fn actions(mut self, sim: SimContext<'_,Shared>) {
		loop {
			for pier in self.piers.iter() {
				// unload the cars
				for car in self.cargo.drain(..) {
					sim.advance(car.load_duration).await;
				}
				
				let begin_loading = sim.now();
				
				// wait until new cars arrive or a timeout occurs
				while self.cargo.len() < self.cargo.capacity() {
					if let Some(car) = sim.select(
						pier.recv(),
						async { sim.advance(self.timeout).await; None }
					).await {
						// a car arrived in time
						sim.shared().car_wait_time.tabulate(sim.now() - car.arrival_time);
						sim.advance(car.load_duration).await;
						self.cargo.push(car);
					} else {
						// the timeout has been triggered
						break;
					}
				}
				
				sim.shared().ferry_load_time.tabulate(sim.now() - begin_loading);
				sim.shared().ferry_cargo_len.tabulate(self.cargo.len() as f64);
				
				// travel to the next harbor
				sim.advance(self.travel_time).await;
			}
		}
	}
}

async fn ferry(sim: SimContext<'_,Shared>, duration: Time, ferries: usize, harbors: usize) {
	let mut ports = Vec::with_capacity(harbors);
	
	// create all of the harbors
	for _ in 0..harbors {
		let (sx, rx) = channel();
		let harbor = Pier {
			rng: SmallRng::from_seed(sim.shared().master_rng.borrow_mut().gen()),
			landing_site: sx
		};
		sim.activate(harbor.actions(sim));
		ports.push(rx);
	}
	
	// create all of the ferries
	for i in 0..ferries {
		let ferry = Ferry {
			cargo: Vec::with_capacity(FERRY_CAPACITY),
			timeout: FERRY_TIMEOUT,
			travel_time: HARBOR_DISTANCE,
			piers: ports.iter().skip(i)
			            .chain(ports.iter().take(i)).cloned().collect()
		};
		sim.activate(ferry.actions(sim));
	}
	
	// await the end of the simulation
	sim.advance(duration).await;
	
	// take cars into account that weren't picked up by a ferry
	for port in ports {
		for _ in 0..port.len() {
			let car = port.recv().await.unwrap();
			sim.shared().car_wait_time.tabulate(sim.now() - car.arrival_time);
		}
	}
}
